#include <fmt/core.h>
#include <fmt/ranges.h>
#include <stdio.h>

#include <CLI/CLI.hpp>
#include <optional>

#include "utils.h"

using json = nlohmann::json;
using std::nullopt;
using std::optional;
using std::string;
using std::vector;

auto tables = ExistTables("./tables.json");

void create_table(const json& j) {
    auto table_name = table_name_of(j);

    if (tables.exists(table_name)) {
        throw std::runtime_error(
            fmt::format("table `{}` exists\n", table_name));
    }
    tables.set(table_name, j["cols"]);
    tables.save();
}

optional<TableCol> get_select_column_define(const vector<string> table_names,
                                            const json& select_col) {
    auto type = type_of(select_col);
    if (type == "identifier") {
        return tables.find_col_in_tables(table_names, select_col["value"]);
    } else if (type == "int" || type == "string" || type == "float") {
        return TableCol{.column_name = select_col["value"].dump(),
                        .type = "const",
                        .data_type = type,
                        .primary_key = false};
    } else if (type == "select_all_column") {
        return TableCol{.column_name = "*",
                        .type = "select_all_column",
                        .data_type = "",
                        .primary_key = false};
    }
    return nullopt;
}

void check_where_clause(const vector<string> table_names, const json& j) {
    if (j.empty()) {
        return;
    }

    auto type = type_of(j);
    auto left = j["left"];
    auto right = j["right"];
    for (const auto curr_branch_name : {"left", "right"}) {
        auto curr_branch = j[curr_branch_name];
        if (left.contains("value")) {
            // 说明到叶节点了
            auto col_def = get_select_column_define(table_names, curr_branch);
            if (!col_def.has_value()) {
                throw std::runtime_error(
                    fmt::format("column `{}` not exists in `{}`\n",
                                curr_branch["value"].dump(2),
                                fmt::join(table_names, ", ")));
            }
        } else {
            check_where_clause(table_names, curr_branch);
        }
    }
}

void check_select_table(const vector<string> table_names,
                        const json& select_cols) {
    for (auto& select_col : select_cols) {
        if (select_col["type"] == "select_all_column") {
            continue;
        }
        auto col_def =
            get_select_column_define(table_names, select_col["target"]);

        if (!col_def.has_value()) {
            throw std::runtime_error(
                fmt::format("column `{}` not exists in `{}`\n",
                            select_col["target"]["value"].dump(2),
                            fmt::join(table_names, ", ")));
        }
    }
}

void process_sql(json& j) {
    let type = j.value("type", "none");

    if (type == "create_stmt") {
        // 创建表
        create_table(j);
    } else if (type == "drop_stmt") {
        if (tables.exists(j["table_name"])) {
            tables.remove(j["table_name"]);
            tables.save();
        }
        return;
    } else if (type == "select_stmt") {
        let table_names = table_names_of(j);
        for (let table_name : table_names) {
            if (!tables.exists(table_name)) {
                fmt::print("error: table `{}` does not exist\n", table_name);
                return;
            }
        }
        check_select_table(table_names, j["select_cols"]);
        check_where_clause(table_names, j["where"]);
    } else {
        throw std::runtime_error(fmt::format(
            "Unknown expression type `{}` total: \n{}", type, j.dump(2)));
    }
}

int main(int argc, char** argv) {
    CLI::App app{"sql 检查器, 附带创建表相关功能"};
    string sql = "", filename = "";

    app.add_option("-f,--file", filename, "输入的 sql 解析后的 json 文件")
        ->check(CLI::ExistingFile);
    app.add_option("-s,--sql", sql, "输入的 sql");

    CLI11_PARSE(app, argc, argv);

    json stmts;
    if (sql.length() > 0) {
        stmts = parse_sql(sql).out;
    } else if (filename.length() > 0) {
        stmts = json::parse(std::ifstream(filename));
    } else {
        std::cout << "请输入 sql 或者 sql 文件" << std::endl;
        return 1;
    }

    for (auto& stmt : stmts) {
        // for(auto& item : stmt.items()) {
        //     fmt::print("{}: {}\n", item.key(), item.value().dump());
        // }
        process_sql(stmt);
    }

    // auto t = res.out[0].value("type", "default");
    // fmt::print("{}\n", t);

    // tables.set("t1", {"a", "b", "c"});
    // tables.set("t3", {"asdge", "safaw", "qwer"});
    // fmt::print("tables.exists {}\n", tables.exists("t1"));
    // tables.save();
    return 0;
}
